# (C) Copyright 2021 by Darren Rowland (advancedthermo.com). All rights reserved.
module AdvancedThermo
using ForwardDiff
using DiffResults
using LsqFit

mutable struct FluidComponent
    Name::String
    Tc::Float64   # critical temperature [K]
    pc::Float64   # critical pressure [Pa]
    ρc::Float64   # critical density [mol/m^3]
    mw::Float64   # molar mass [kg/mol]
    ω::Float64    # acentric factor
    # ideal-gas terms
    ni::Vector{Float64}
    vi::Vector{Float64}
    # polynomial-type parameters
    np::Vector{Float64}
    dp::Vector{Int64}
    tp::Vector{Float64}
    # exponential-type parameters
    ne::Vector{Float64}
    de::Vector{Int64}
    te::Vector{Float64}
    γe::Vector{Float64}
    ce::Vector{Int64}
end
mutable struct FluidMixture
    components::Vector{FluidComponent}
    Tc::Vector{Float64}
    ρc::Vector{Float64}
    βv::Matrix{Float64}
    γv::Matrix{Float64}
    βT::Matrix{Float64}
    γT::Matrix{Float64}
end
const R = 8.314472

################
#  PROPERTIES  #
################

function ρ(m::FluidMixture,p,T,x)
    ρ_lo = 1.e-6
    (p_lo,∂p∂ρ,∂²p∂ρ²) = derivatives_of_p_wrt_ρ(m,ρ_lo,T,x)
    nit = 0
    while abs(p_lo-p)/p > 1.e-10 && nit < 30
        dρ = (p-p_lo)/∂p∂ρ
        dρ_maximum = -∂p∂ρ/∂²p∂ρ² > 0 ? -∂p∂ρ/∂²p∂ρ² : 1e100
        if ∂p∂ρ > 0.0 && dρ < dρ_maximum
            if ρ_lo + dρ > 0.0
                ρ_lo += dρ
                (p_lo,∂p∂ρ,∂²p∂ρ²) = derivatives_of_p_wrt_ρ(m,ρ_lo,T,x)
            else
                ρ_lo = -1.0
                p_lo = p
            end
        else
            ρ_lo = -1.0
            p_lo = p
        end
        nit += 1
    end
    ρ_hi = 3.e5
    (p_hi,∂p∂ρ,∂²p∂ρ²) = derivatives_of_p_wrt_ρ(m,ρ_hi,T,x)
    nit = 0
    while abs(p_hi-p)/p > 1.e-10 && nit < 30
        dρ = (p-p_hi)/∂p∂ρ
        dρ_minimum = -∂p∂ρ/∂²p∂ρ²
        if dρ>dρ_minimum
            ρ_hi += dρ
            (p_hi,∂p∂ρ,∂²p∂ρ²) = derivatives_of_p_wrt_ρ(m,ρ_hi,T,x)
        else
            ρ_hi = -1.0
            p_hi = p
        end
        nit += 1
    end
    if ρ_lo>=0.0 && ρ_hi>0.0
        g_lo = g(m,ρ_lo,T,x)
        g_hi = g(m,ρ_hi,T,x)
        if g_lo < g_hi
            return ρ_lo
        else
            return ρ_hi
        end
    elseif ρ_lo >= 0.0
        return ρ_lo
    elseif ρ_hi > 0.0
        return ρ_hi
    else
        # couldn't locate either root, attempt bisection search
        ρ_lo = 1e-3
        (p_lo, _) = derivatives_of_p_wrt_ρ(m,ρ_lo,T,x)
        ρ_hi = 3e5
        (p_hi, _) = derivatives_of_p_wrt_ρ(m,ρ_hi,T,x)
        ρ_mid = 0.5*(ρ_lo + ρ_hi)
        while ρ_hi - ρ_lo > 1e-8
            ρ_mid = 0.5*(ρ_lo + ρ_hi)
            (p_mid, _) = derivatives_of_p_wrt_ρ(m,ρ_mid,T,x)
            if p_mid < p
                ρ_lo = ρ_mid
                p_lo = p_mid
            else
                ρ_hi = ρ_mid
                p_hi = p_mid
            end
        end
        return ρ_mid
    end
end
function a(m::FluidMixture,v,T,x)
    return R*T*α(m,inv(v),T,x)
end
function ∂a∂T(m::FluidMixture,v,T,x)
    f = T -> a(m,v,T,x)
    return ForwardDiff.derivative(f,T)
end
function ∂a∂v(m::FluidMixture,v,T,x)
    f = v -> a(m,v,T,x)
    return ForwardDiff.derivative(f,v)
end
function s(m::FluidMixture,ρ,T,x)
    return -∂a∂T(m,inv(ρ),T,x)
end
function p(m::FluidMixture,ρ,T,x)
    return -∂a∂v(m,inv(ρ),T,x)
end
function p(m::FluidMixture,VTn)
    ntotal = sum(VTn[3:end])
    return p(m,ntotal/VTn[1],VTn[2],VTn[3:end]./ntotal)
end
function Z(m::FluidMixture,VTn)
    ntotal = sum(VTn[3:end])
    return Z(m,ntotal/VTn[1],VTn[2],VTn[3:end]./ntotal)
end
function Z(m::FluidMixture,ρ,T,x)
    return p(m::FluidMixture,ρ,T,x)/(ρ*R*T)
end
function g(m::FluidMixture,ρ,T,x)
    v = inv(ρ)
    return a(m,v,T,x) + p(m,ρ,T,x)*v
end
function lnϕ(m::FluidMixture,ρ,T,x)
    result = DiffResults.GradientResult(x)
    f = x -> nαʳ(m,vcat(inv(ρ),T,x))
    result = ForwardDiff.gradient!(result, f, x)
    return DiffResults.gradient(result) .- log(Z(m,ρ,T,x))
end
function α(m::FluidMixture,ρ,T,x)
    return αᵒ(m,ρ,T,x) + αʳ(m,ρ,T,x)
end
function αᵒ(fluid::FluidComponent,ρ,T)
    δᵢ = ρ/fluid.ρc
    τᵢ = fluid.Tc/T
    s = [
        1.0, 
        τᵢ, 
        log(τᵢ), 
        log(abs(sinh(fluid.vi[4]*τᵢ))), 
       -log(abs(cosh(fluid.vi[5]*τᵢ))), 
        log(abs(sinh(fluid.vi[6]*τᵢ))), 
       -log(abs(cosh(fluid.vi[7]*τᵢ)))]
    return log(δᵢ) + fluid.ni's
end
function αᵒ(m::FluidMixture,ρ,T,x)
    result = 0.0
    for i = 1:length(x)
        if x[i]>0.0
            result += x[i]*(αᵒ(m.components[i],ρ,T)+log(x[i]))
        end
    end
    return result
end
function αʳ(fluid::FluidComponent,ρ,T)
    δ = ρ/fluid.ρc
    τ = fluid.Tc/T
    return αʳ(δ,τ,fluid)
end
function αʳ(m::FluidMixture,ρ,T,x)
    δ = ρ/ρᵣ(x,m.ρc,m.βv,m.γv)
    τ = Tᵣ(x,m.Tc,m.βT,m.γT)/T
    return αʳ(δ,τ,x,m)
end
function nαʳ(m::FluidMixture,VTn)
    ntotal = sum(VTn[3:end])
    return ntotal*αʳ(m,ntotal/VTn[1],VTn[2],VTn[3:end]./ntotal)
end
function αʳ(δ,τ,x,m::FluidMixture)
    return sum(x.*αʳ.(δ,τ,m.components))
end
function αʳ(δ,τ,f::FluidComponent)
    return αʳ(δ,τ,f.np,f.dp,f.tp) + αʳ(δ,τ,f.ne,f.de,f.te,f.γe,f.ce)
end
function αʳ(δ,τ,n,d,t)
    return sum(@.(n*δ^d*τ^t))
end
function αʳ(δ,τ,n,d,t,γ,c)
    return sum(@.(n*δ^d*τ^t*exp(-γ*δ^c)))
end
function ρᵣ(x,ρc,βv,γv)
    return inv(vᵣ(x,inv.(ρc),βv,γv))
end
function vᵣ(x,vc,βv,γv)
    result = 0.0
    for j = 1:length(x)
        if x[j]>0.0
            for i = 1:length(x)
                if x[i]>0.0
                    vᵢⱼ = (vc[i]^(1/3)+vc[j]^(1/3))^3/8.0
                    result += ( x[i]*x[j]*βv[i,j]*γv[i,j]*
                                (x[i]+x[j])/(βv[i,j]^2*x[i]+x[j])*vᵢⱼ )
                end
            end
        end
    end
    return result
end
function Tᵣ(x,Tc,βT,γT)
    result = 0.0
    for j = 1:length(x)
        if x[j]>0.0
            for i = 1:length(x)
                if x[i]>0.0
                    Tᵢⱼ = sqrt(Tc[i]*Tc[j])
                    result += ( x[i]*x[j]*βT[i,j]*γT[i,j]*
                                (x[i]+x[j])/(βT[i,j]^2*x[i]+x[j])*Tᵢⱼ )
                end
            end
        end
    end
    return result
end

#################
#  DERIVATIVES  #
#################

function ∂lnϕ∂n(m::FluidMixture,V,T,n)
    dna = AdvancedThermo.derivatives_of_nαʳ(m,V,T,n)
    dp = AdvancedThermo.derivatives_of_p(m,V,T,n)
    ∂p∂V = DiffResults.gradient(dp)[1]
    ∂p∂n = DiffResults.gradient(dp)[3:end]
    ∂lnϕ∂n = DiffResults.hessian(dna)[3:end,3:end] .+ 1 .+
                  ∂p∂n*∂p∂n'./(R*T*∂p∂V)
    return ∂lnϕ∂n
end
function derivatives_of_a(m::FluidMixture,ρ,T,x)
    z = vcat(inv(ρ),T,x)
    result = DiffResults.HessianResult(z)
    f = z -> a(m,z[1],z[2],z[3:end])
    result = ForwardDiff.hessian!(result, f, z)
    return result
end
function derivatives_of_nαʳ(m::FluidMixture,V,T,n)
    z = vcat(V,T,n)
    result = DiffResults.HessianResult(z)
    f = z -> nαʳ(m,z)
    result = ForwardDiff.hessian!(result, f, z)
    return result
end
function derivatives_of_p_wrt_ρ(m::FluidMixture,ρ,T,x)
    result = DiffResults.HessianResult([ρ])
    f = ρ -> p(m,ρ[1],T,x)
    result = ForwardDiff.hessian!(result, f, [ρ])
    pcalc = DiffResults.value(result)
    ∂p∂ρ = DiffResults.gradient(result)[1]
    ∂²p∂ρ² = DiffResults.hessian(result)[1,1]
    return (pcalc,∂p∂ρ,∂²p∂ρ²)
end
function derivatives_of_p(m::FluidMixture,V,T,n)
    z = vcat(V,T,n)
    result = DiffResults.GradientResult(z)
    f = z -> p(m,z)
    result = ForwardDiff.gradient!(result, f, z)
    return result
end
function derivatives_of_Z(m::FluidMixture,V,T,n)
    z = vcat(V,T,n)
    result = DiffResults.HessianResult(z)
    f = z -> Z(m,z)
    result = ForwardDiff.hessian!(result, f, z)
    return result
end
using CSV
function calculate_goodness_of_fit(Tc,ρc)
    methane = AdvancedThermo.get_fluid_component("methane")
    helium = AdvancedThermo.get_fluid_component("helium")
    # Dummy value for critical pressure
    helium = AdvancedThermo.new_critical!(helium,1e6,Tc,ρc)
    m = AdvancedThermo.create_mixture([methane,helium])
    (γv12, γT12) = AdvancedThermo.arithmetic_γ(methane, helium)
    m.γv[1,2] = m.γv[2,1] = γv12
    m.γT[1,2] = m.γT[2,1] = γT12
    file = "test/pTxy_methane_helium.csv"
    ssq = 0.0
    for row in CSV.File(file)
        T = row.T
        p = row.p*1e3
        x = [row.x1, row.x2]
        lnf_liq = log.(x)+AdvancedThermo.lnϕ(m,AdvancedThermo.ρ(m,p,T,x),T,x)
        y = [row.y1, row.y2]
        lnf_vap = log.(y)+AdvancedThermo.lnϕ(m,AdvancedThermo.ρ(m,p,T,y),T,y)
        #@info lnf_liq, lnf_vap
        ssq += sum((lnf_liq.-lnf_vap).^2)
    end
    rmsd = sqrt(ssq/2/128)
    return (ssq, rmsd)
end
function calculate_goodness_of_fit_flash(Tc,ρc)
    methane = AdvancedThermo.get_fluid_component("methane")
    helium = AdvancedThermo.get_fluid_component("helium")
    # Dummy value for critical pressure
    helium = AdvancedThermo.new_critical!(helium,1e6,Tc,ρc)
    m = AdvancedThermo.create_mixture([methane,helium])
    (γv12, γT12) = AdvancedThermo.arithmetic_γ(methane, helium)
    m.γv[1,2] = m.γv[2,1] = γv12
    m.γT[1,2] = m.γT[2,1] = γT12
    file = "test/pTxy_methane_helium.csv"
    ssq = 0.0
    devK1 = 0.0
    devK2 = 0.0
    for row in CSV.File(file)
        z = [0.5*(row.x1+row.y1), 0.5*(row.x2+row.y2)]
        try 
            (x,y,β) = flash_with_stability(m,row.p*1e3,row.T,z)
            if β==0.0 || β==1.0
                # penalise single phase results
                devK1 = 10.0
                devK2 = 10.0
            else
                devK1 = log(row.y1/row.x1) - log(y[1]/x[1])
                devK2 = log(row.y2/row.y1) - log(y[2]/x[2])
            end
        catch e
            # penalise other errors
            devK1 = 10.0
            devK2 = 10.0
        end
        ssq += devK1^2 + devK2^2
    end
    rmsd = sqrt(ssq/2/128)
    return (ssq, rmsd)
end
function calculate_statistics(dev)
    n = length(dev)
    AAD = sum(abs.(dev))/n
    Bias = sum(dev)/n
    RMS = sqrt(sum(dev.^2)/n)
    Max = maximum(dev)>-minimum(dev) ? maximum(dev) : minimum(dev)
    return (AAD, Bias, RMS, Max)
end
function flash_statistics(file, N, mixture)
    devK1 = zeros(N)
    devK2 = zeros(N)
    K1 = zeros(N)
    K2 = zeros(N)
    i = 0
    for row in CSV.File(file)
        z = [0.5*(row.x1+row.y1), 0.5*(row.x2+row.y2)]
        try 
            (x,y,β) = AdvancedThermo.flash_with_stability(mixture,row.p*1e3,row.T,z)
            if β==0.0 || β==1.0
                @info "Single phase - $(row.Reference) datum $(row.n)"
            else
                i += 1
                devK1[i] = log(row.y1/row.x1) - log(y[1]/x[1])
                devK2[i] = log(row.y2/row.y1) - log(y[2]/x[2])
                K1[i] = log(row.y1/row.x1)
                K2[i] = log(row.y2/row.x2)
                #@info "K1=$(row.y1/row.x1) K2=$(row.y2/row.y1); y1=$(y[1]) x1=$(x[1]) K1=$(y[1]/x[1]) K2=$(y[2]/x[2])"
            end
        catch e
            @info "Internal error - $(row.Reference) datum $(row.n)"
            println(e.msg)
        end
    end
    (AADK1,BiasK1,RMSK1,MaxK1) = calculate_statistics(devK1[1:i])
    (AADK2,BiasK2,RMSK2,MaxK2) = calculate_statistics(devK2[1:i])
    println("Statistic      lnK(methane)               lnK(helium)")
    println("AAD            $AADK1               $AADK2")
    println("Bias           $BiasK1              $BiasK2")
    println("RMS            $RMSK1               $RMSK2")
    println("Max            $MaxK1               $MaxK2")
    return (K1[1:i], K2[1:i], devK1[1:i], devK2[1:i])
end
function fit_critical_constants(file, mixture, T0, ρ0)
    function model(x,p)
        @info p
        mixture.components[2] = new_critical!(mixture.components[2], 1.e6, p[1], p[2]*1e3)
        (γv, γT) = arithmetic_γ(mixture.components[1], mixture.components[2])
        mixture.γv[1,2] = mixture.γv[2,1] = γv
        mixture.γT[1,2] = mixture.γT[2,1] = γT
        devK1 = zeros(128)
        devK2 = zeros(128)
        i = 0
        for row in CSV.File(file)
            i += 1
            T = row.T
            p = row.p*1e3
            x = [row.x1, row.x2]
            lnf_liq = log.(x)+AdvancedThermo.lnϕ(mixture,AdvancedThermo.ρ(mixture,p,T,x),T,x)
            y = [row.y1, row.y2]
            lnf_vap = log.(y)+AdvancedThermo.lnϕ(mixture,AdvancedThermo.ρ(mixture,p,T,y),T,y)
            devK1[i] = lnf_liq[1] - lnf_vap[1]
            devK2[i] = lnf_liq[2] - lnf_vap[2]
        end
        return vcat(devK1[1:i], devK2[1:i])
    end
    p0 = [T0, ρ0/1000]
    fit = LsqFit.curve_fit(model, zeros(256), zeros(256), p0)
    return coef(fit)
end
function fit_critical_constants_flash(file, mixture)
    function model(x,p)
        mixture.components[2] = new_critical!(mixture.components[2], 1.e6, p[1], p[2]*1e3)
        (γv, γT) = arithmetic_γ(mixture.components[1], mixture.components[2])
        mixture.γv[1,2] = mixture.γv[2,1] = γv
        mixture.γT[1,2] = mixture.γT[2,1] = γT
        devK1 = zeros(128)
        devK2 = zeros(128)
        i = 0
        for row in CSV.File(file)
            i += 1
            z = [0.5*(row.x1+row.y1), 0.5*(row.x2+row.y2)]
            try 
                (x,y,β) = flash_with_stability(mixture,row.p*1e3,row.T,z)
                if β==0.0 || β==1.0
                    # penalise single phase results
                    devK1[i] = 10.0
                    devK2[i] = 10.0
                else
                    devK1[i] = log(row.y1/row.x1) - log(y[1]/x[1])
                    devK2[i] = log(row.y2/row.y1) - log(y[2]/x[2])
                end
            catch e
                # penalise other errors
                devK1[i] = 10.0
                devK2[i] = 10.0
            end
        end
        return vcat(devK1[1:i], devK2[1:i])
    end
    p0 = [13.0, 28.0]
    fit = LsqFit.curve_fit(model, zeros(256), zeros(256), p0)
    return coef(fit)
end

######################
#  PHASE EQUILIBRIA  #
######################

function flash(m::FluidMixture,p,T,z)
    lnK = lnK_Wilson(m,p,T)
    K = exp.(lnK)
    (x,y,β) = flash_initial_K(m,p,T,z,K)
    return (x,y,β)
end
function flash_with_stability(m::FluidMixture,p,T,z)
    (W, K, tpd, tm) = stability_helium_binary(m,p,T,z)
    if tpd < -1.e-12
        (x,y,β) = flash_initial_K(m,p,T,z,K)
    else
        x = z
        y = z
        β = 0.0
    end
    return (x,y,β)
end
function flash_helium_binary(m::FluidMixture,p,T,z)
    K = exp.(lnK_helium_binary(m::FluidMixture,p,T))
    return flash_initial_K(m,p,T,z,K)
end
function flash_initial_K(m::FluidMixture,p,T,z,K)
    β_prev = -1.0
    (x,y,β) = solve_xyβ(z,K)
    nit = 0
    while abs(β-β_prev)>1.e-9 && nit<100
        nit = nit + 1
        lnϕ_liq = lnϕ(m,ρ(m,p,T,x),T,x)
        lnϕ_vap = lnϕ(m,ρ(m,p,T,y),T,y)
        β_prev = β
        (x,y,β) = solve_xyβ(z,exp.(lnϕ_liq.-lnϕ_vap))
    end
    return (x,y,β)
end
function solve_xyβ(z,K)
    (f0,df0) = RachfordRice(z,K,0.0)
    (f1,df1) = RachfordRice(z,K,1.0)
    if f0 > 0.0 && f1 < 0.0
        β = 0.5
        (f,df) = RachfordRice(z,K,β)
        nit = 0
        while abs(f)>1.e-12 && nit<100
            nit += 1
            β -= f/df
            (f,df) = RachfordRice(z,K,β)
        end
        x = @.(z/(1+β*(K-1)))
        y = K.*x
        return (x,y,β)
    else
        return (z,z,1.0)
    end
end
function RachfordRice(z,K,β)
    s = @.((K-1)/(1+β*(K-1)))
    return (z's, -z's.^2)
end
function stability_helium_binary(m::FluidMixture,p,T,z)
    lnK = lnK_helium_binary(m,p,T)
    K = exp.(lnK)
    W = K.*z
    Wv = minimise_tm(m,p,T,z,W)
    Kv = Wv./z
    W = z./K
    Wl = minimise_tm(m,p,T,z,W)
    Kl = z./Wl
    tpd_v = tpd(m,p,T,z,Wv./sum(Wv))
    tpd_l = tpd(m,p,T,z,Wl./sum(Wl))
    tm_v = tm(m,p,T,z,Wv)
    tm_l = tm(m,p,T,z,Wl)
    if tpd_v < tpd_l
        return (Wv, Kv, tpd_v, tm_v)
    else
        return (Wl, Kl, tpd_l, tm_l)
    end
end
function minimise_tm(m::FluidMixture,p,T,z,W)
    d = log.(z) .+ lnϕ(m,ρ(m,p,T,z),T,z)
    tm_ = tm(m,p,T,z,W)
    for i = 1:400
        w = W./sum(W)
        # Michelsen/Mollerup (10.51)
        Wnew = exp.(d .- lnϕ(m,ρ(m,p,T,w),T,w))
        tmnew = tm(m,p,T,z,Wnew)
        Wnew, Wnew./sum(Wnew)
        if maximum(abs.(Wnew.-W))<1.e-5
            return Wnew
        end
        # reject steps that don't decrease tm
        if tmnew >= tm_
            return W
        end
        W .= Wnew
        tm_ = tmnew
    end
    @warn "failed to minimise tm(W): p=$p, T=$T, z=$z, W=$W"
    return W
end
function tpd(m::FluidMixture,p,T,z,w)
    # Michelsen/Mollerup (9.4) and (10.42)
    return sum(w.*(log.(w).+lnϕ(m,ρ(m,p,T,w),T,w).-log.(z).-lnϕ(m,ρ(m,p,T,z),T,z)))
end
function tm(m::FluidMixture,p,T,z,W)
    WT = sum(W)
    w = W./WT
    return 1 - WT + WT*log(WT) + WT*tpd(m,p,T,z,w)
end
function lnK_Wilson(m::FluidMixture,p,T)
    n = length(m.components)
    lnK = zeros(n)
    for i = 1:n
        lnK[i] = log(m.components[i].pc/p) + 5.37*
                 (1+m.components[i].ω)*(1-m.components[i].Tc/T)
    end
    return lnK
end
function lnK_helium_binary(m::FluidMixture,p,T)
    x1 = [1.0, 0.0]
    x2 = [0.0, 1.0]
    ρ1 = ρ(m,p,T,x1)
    ρ2 = ρ(m,p,T,x2)
    lnϕ_1 = lnϕ(m,ρ1,T,x1)
    lnϕ_2 = lnϕ(m,ρ2,T,x2)
    lnK = ρ1>ρ2 ? lnϕ_1.-lnϕ_2 : lnϕ_2.-lnϕ_1
    return lnK
end

######################
#  MODEL PARAMETERS  #
######################

function create_mixture(fluids::Vector{FluidComponent})
    n = length(fluids)
    Tc = zeros(n)
    ρc = zeros(n)
    for i = 1:n
        Tc[i] = fluids[i].Tc
        ρc[i] = fluids[i].ρc
    end
    (βv, γv, βT, γT) = get_interaction_matrices(fluids)
    return FluidMixture(fluids, Tc, ρc, βv, γv, βT, γT)
end
function new_critical!(fluid::FluidComponent,pc,Tc,ρc)
    ρr = ρc/fluid.ρc
    Tr = fluid.Tc/Tc
    fluid.pc = pc
    fluid.Tc = Tc
    fluid.ρc = ρc
    # ideal contribution
    fluid.ni[1] += log(ρr) + fluid.ni[3]*log(Tr)
    fluid.ni[2] *= Tr
    @. fluid.vi *= Tr
    # residual contribution
    @. fluid.np *= ρr^fluid.dp * Tr^fluid.tp
    @. fluid.ne *= ρr^fluid.de * Tr^fluid.te
    @. fluid.γe *= ρr^fluid.ce
    return fluid
end
function arithmetic_γ(fluid1::FluidComponent, fluid2::FluidComponent)
    v1 = inv(fluid1.ρc)
    v2 = inv(fluid2.ρc)
    T1 = fluid1.Tc
    T2 = fluid2.Tc
    return arithmetic_γ(v1,T1,v2,T2)
end
function arithmetic_γ(v1,T1,v2,T2)
    # Kunz & Wagner (2012) Eq. (39)
    γv = 4.0*(v1+v2)/(v1^(1/3)+v2^(1/3))^3
    γT = 0.5*(T1+T2)/sqrt(T1*T2)
    return (γv,γT)
end
function get_interaction_matrices(fluids::Vector{FluidComponent})
    n = length(fluids)
    βv = ones(n,n)
    γv = ones(n,n)
    βT = ones(n,n)
    γT = ones(n,n)
    n = length(fluids)
    # Parameters from Kunz and Wagner (2012), Table A8
    for j = 1:n
        for i = 1:n
            if fluids[i].Name=="methane" && fluids[j].Name=="helium"
                # only γv and γT non-unity
                γv[i,j] = γv[j,i] = 0.881405683
                γT[i,j] = γT[j,i] = 3.159776855
            elseif fluids[i].Name=="nitrogen" && fluids[j].Name=="helium"
                βv[i,j] = 0.969501055
                βv[j,i] = inv(βv[i,j])
                γv[i,j] = γv[j,i] = 0.932629867
                βT[i,j] = 0.692868765
                βT[j,i] = inv(βT[i,j])
                γT[i,j] = γT[j,i] = 1.471831580
            else
                # TODO(DR): more interactions as needed
            end
        end
    end
    return (βv, γv, βT, γT)
end
function get_fluid_component(name)
    if name=="methane"
        # refer Kunz & Wagner (2012) Table A1
        ni = [
            19.597508817
            -83.959667892
            3.00088
            0.76315
            0.00460
            8.74432
            -4.46921
        ]
        vi = [
            0.0
            0.0
            0.0
            4.306474465
            0.936220902
            5.577233895
            5.722644361
        ]
        # refer Kunz & Wagner (2012) Tables A3 and A5
        np = [
            0.57335704239162
            -0.16760687523730e1
            0.23405291834916
            -0.21947376343441
            0.16369201404128e-1
            0.15004406389280e-1
        ]
        dp = [
            1
            1
            2
            2
            4
            4
        ]
        tp = [
            0.125
            1.125
            0.375
            1.125
            0.625
            1.500
        ]
        ne = [
            0.98990489492918e-1
            0.58382770929055
            -0.74786867560390
            0.30033302857974
            0.20985543806568
            -0.18590151133061e-1
            -0.15782558339049
            0.12716735220791
            -0.32019743894346e-1
            -0.68049729364536e-1
            0.24291412853736e-1
            0.51440451639444e-2
            -0.19084949733532e-1
            0.55229677241291e-2
            -0.44197392976085e-2
            0.40061416708429e-1
            -0.33752085907575e-1
            -0.25127658213357e-2
        ]
        de = [
            1
            1
            1
            2
            3
            6
            2
            3
            3
            4
            4
            2
            3
            4
            5
            6
            6
            7
        ]
        te = [
            0.625
            2.625
            2.750
            2.125
            2.000
            1.750
            4.500
            4.750
            5.000
            4.000
            4.500
            7.500
            14.000
            11.500
            26.000
            28.000
            30.000
            16.000
        ]
        ce = [
            1
            1
            1
            1
            1
            1
            2
            2
            2
            2
            2
            3
            3
            3
            6
            6
            6
            6
        ]
        return FluidComponent(
            name,
            190.564,
            4.5992e6,
            10139.342719,
            16.04246e-3,
            0.01142,
            ni,
            vi,
            np,
            dp,
            tp,
            ne,
            de,
            te,
            ones(length(te)),
            ce
        )
    elseif name=="nitrogen"
        # refer Kunz & Wagner (2012) Table A1
        ni = [
            11.083407489
            -22.202102428
            2.50031
            0.13732
            -0.14660
            0.90066
            0.0
        ]
        vi = [
            0.0
            0.0
            0.0
            5.251822620
            -5.393067706
            13.788988208
            1.0
        ]
        # refer Kunz & Wagner (2012) Tables A3 and A5
        np = [
            0.59889711801201
            -0.16941557480731e1
            0.24579736191718
            -0.23722456755175
            0.17954918715141e-1
            0.14592875720215e-1
        ]
        dp = [
            1
            1
            2
            2
            4
            4
        ]
        tp = [
            0.125
            1.125
            0.375
            1.125
            0.625
            1.500
        ]
        ne = [
            0.10008065936206
            0.73157115385532
            -0.88372272336366
            0.31887660246708
            0.20766491728799
            -0.19379315454158e-1
            -0.16936641554983
            0.13546846041701
            -0.33066712095307e-1
            -0.60690817018557e-1
            0.12797548292871e-1
            0.58743664107299e-2
            -0.18451951971969e-1
            0.47226622042472e-2
            -0.52024079680599e-2
            0.43563505956635e-1
            -0.36251690750939e-1
            -0.28974026866543e-2
        ]
        de = [
            1
            1
            1
            2
            3
            6
            2
            3
            3
            4
            4
            2
            3
            4
            5
            6
            6
            7
        ]
        te = [
            0.625
            2.625
            2.750
            2.125
            2.000
            1.750
            4.500
            4.750
            5.000
            4.000
            4.500
            7.500
            14.000
            11.500
            26.000
            28.000
            30.000
            16.000
        ]
        ce = [
            1
            1
            1
            1
            1
            1
            2
            2
            2
            2
            2
            3
            3
            3
            6
            6
            6
            6
        ]
        return FluidComponent(
            name,
            126.192,
            3.3958e6,
            11183.9,
            28.0134e-3,
            0.0373,
            ni,
            vi,
            np,
            dp,
            tp,
            ne,
            de,
            te,
            ones(length(te)),
            ce
        )
    elseif name=="helium"
        # refer Kunz & Wagner (2012) Table A1
        ni = [
            13.628409737
            -143.470759602
            1.50000
            0.0
            0.0
            0.0
            0.0
        ]
        vi = [
            0.0
            0.0
            0.0
            1.0
            1.0
            1.0
            1.0
        ]
        # refer Kunz & Wagner (2012) Tables A4-A5
        np = [
            -0.45579024006737
            0.12516390754925e1
            -0.15438231650621e1
            0.20467489707221e-1
        ]
        dp = [
            1
            1
            1
            4
        ]
        tp = [
            0.000
            0.125
            0.750
            1.000
        ]
        ne = [
            -0.34476212380781
            -0.20858459512787e-1
            0.16227414711778e-1
            -0.57471818200892e-1
            0.19462416430715e-1
            -0.33295680123020e-1
            -0.10863577372367e-1
            -0.22173365245954e-1
        ]
        de = [
            1
            3
            5
            5
            5
            2
            1
            2
        ]
        te = [
            0.750
            2.625
            0.125
            1.250
            2.000
            1.000
            4.500
            5.000
        ]
        ce = [
            1
            1
            1
            1
            1
            2
            3
            3
        ]
        return FluidComponent(
            name,
            5.1953,
            0.226e6,
            17399.0,
            4.002602e-3,
            -0.395,
            ni,
            vi,
            np,
            dp,
            tp,
            ne,
            de,
            te,
            ones(length(te)),
            ce
        )
    elseif name=="argon"
        # refer Kunz & Wagner (2012) Table A1
        ni = [
            8.316631500
            -4.946502600
            1.50000
            0.0
            0.0
            0.0
            0.0
        ]
        vi = [
            0.0
            0.0
            0.0
            1.0
            1.0
            1.0
            1.0
        ]
        # refer Kunz & Wagner (2012) Tables A4-A5
        np = [
            0.85095714803969
            -0.24003222943480e1
            0.54127841476466
            0.16919770692538e-1
            0.68825965019035e-1
            0.21428032815338e-3
        ]
        dp = [
            1
            1
            1
            2
            3
            7
        ]
        tp = [
            0.250
            1.125
            1.500
            1.375
            0.250
            0.875
        ]
        ne = [
            0.17429895321992
            -0.33654495604194e-1
            -0.13526799857691
            -0.16387350791552e-1
            -0.24987666851475e-1
            0.88769204815709e-2
        ]
        de = [
            2
            5
            1
            4
            3
            4
        ]
        te = [
            0.625
            1.750
            3.625
            3.625
            14.500
            12.000
        ]
        ce = [
            1
            1
            2
            2
            3
            3
        ]
        return FluidComponent(
            name,
            150.687,
            4.8798e6,
            13407.429659,
            39.948e-3,
            0.0,
            ni,
            vi,
            np,
            dp,
            tp,
            ne,
            de,
            te,
            ones(length(te)),
            ce
        )
    else
        # TODO(DR): more fluids as needed
    end
end
end # module
